
/* This file is part of KDevelop
    Copyright 2007 David Nolden <david.nolden.kdevelop@art-master.de>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/


#include "templatedeclaration.h"

#include <qatomic.h>
#include <kglobal.h>

#include <duchain/declaration.h>
#include <duchain/forwarddeclaration.h>

#include "templateparameterdeclaration.h"
#include "cppducontext.h"
#include "expressionparser.h"

using namespace KDevelop;
using namespace Cpp;


QMutex TemplateDeclaration::instantiationsMutex(QMutex::Recursive);

typedef CppDUContext<KDevelop::DUContext> StandardCppDUContext;

struct AtomicIncrementer {
#if QT_VERSION < 0x040400
  typedef QBasicAtomic Int;
#else
  typedef QBasicAtomicInt Int;
#endif
  AtomicIncrementer( Int *cnt ) : c(cnt) {
    c->ref();
  }

  ~AtomicIncrementer() {
    c->deref();
  }

  Int* c;
};

uint qHash( const ExpressionEvaluationResult& key )
{
  uint ret = 13*key.instance.isInstance;// + 17*(quint64)key.instance.declaration*/;

/*  if( key.type )
    ret += qHash(key.type->toString()); ///@todo Expensive
  else*/
  ret += key.identifier().hash();
/*  if( key.type )
    return qHash(key.type);*/
  
  return ret;
}

uint qHash( const QList<ExpressionEvaluationResult>& key ) {
  uint ret = 0;
  foreach( const ExpressionEvaluationResult& expr, key )
    ret = qHash(expr) + ret * 7;
  return ret;
}

bool InstantiationKey::operator==(const InstantiationKey & rhs) const
{
  return qHash(args) == qHash(rhs.args);
}

uint qHash( const InstantiationKey& key ) {
  return qHash( key.args );
}

AbstractType* applyPointerReference( AbstractType* ptr, const KDevelop::TypeIdentifier& id ) {
  AbstractType* ret = ptr;
  for( int a = 0; a < id.pointerDepth(); ++a ) {
    Declaration::CVSpec spec = Declaration::CVNone;
    if( id.isConstPointer( a ) )
      spec = Declaration::Const;
    
    CppPointerType* newRet = new CppPointerType( spec );
    newRet->setBaseType( AbstractType::Ptr(ret) );
    ret = newRet;
  }

  if(id.isReference() ) {
    Declaration::CVSpec spec = Declaration::CVNone;
    if( id.isConstant() )
      spec = Declaration::Const;
    
    CppReferenceType* newRet = new CppReferenceType( spec );
    newRet->setBaseType( AbstractType::Ptr(ret) );
    ret = newRet;
  }

  return ret;
}

///Finds out whether any DelayedType's are involved in a given type(It searches return-values, argument-types, base-classes, etc.)
struct DelayedTypeSearcher : public KDevelop::SimpleTypeVisitor {
  bool found;

  DelayedTypeSearcher() : found(false) {
  }

  virtual bool visit ( const AbstractType* type ) {
    if( found ) return false;
    if( dynamic_cast<const DelayedType*>(type) )
      found = true;
    if( const CppClassType* classType = dynamic_cast<const CppClassType*>(type) ) {
        foreach( const CppClassType::BaseClassInstance& base, classType->baseClasses() )
          visit( base.baseClass.data() );
    }
    return !found;
  }

  virtual bool visit (const StructureType *) {
    ///We do not want to visit member-types, so return false here
    return false;
  }
};

/**
 * Returns whether any count of reference/pointer-types are followed by a delayed type
 * */
const DelayedType* containsDelayedType(const AbstractType* type)
{
  const PointerType* pType = dynamic_cast<const PointerType*>(type);
  const ReferenceType* rType = dynamic_cast<const ReferenceType*>(type);
  const DelayedType* delayedType = dynamic_cast<const DelayedType*>(type);
  const CppClassType* classType = dynamic_cast<const CppClassType*>(type);
  if( pType )
    return containsDelayedType(pType->baseType().data());
  if( rType )
    return containsDelayedType(rType->baseType().data());
  if( classType ) {
    foreach( const CppClassType::BaseClassInstance& base, classType->baseClasses() ) {
      const DelayedType* delayed = dynamic_cast<const DelayedType*>(base.baseClass.data());
      if( delayed )
        return delayed;
      else if( (delayed = containsDelayedType( base.baseClass.data() )) )
        return delayed;
    }
  }
  return delayedType;
}

///Replaces any DelayedType's in interesting positions with their resolved versions, if they can be resolved.
struct DelayedTypeResolver : public KDevelop::TypeExchanger {
  const KDevelop::DUContext* searchContext;
  
  const KDevelop::ImportTrace& inclusionTrace;
  static AtomicIncrementer::Int depth_counter;
  KDevelop::DUContext::SearchFlags searchFlags;

  DelayedTypeResolver(const KDevelop::DUContext* _searchContext, const KDevelop::ImportTrace& _inclusionTrace, KDevelop::DUContext::SearchFlags _searchFlags = KDevelop::DUContext::NoUndefinedTemplateParams) : searchContext(_searchContext), inclusionTrace(_inclusionTrace), searchFlags(_searchFlags) {
  }

  virtual AbstractType* exchange( const AbstractType* type )
  {
    AtomicIncrementer inc(&depth_counter);
    if( depth_counter > 20 ) {
      kDebug(9007) << "Too much depth in DelayedTypeResolver::exchange, while exchanging" << (type ? type->toString() : QString("(null)"));
      return const_cast<AbstractType*>(type); ///@todo remove const_cast;
    }
    const DelayedType* delayedType = dynamic_cast<const DelayedType*>(type);

    if( delayedType && delayedType->kind() == DelayedType::Delayed ) {
      if( !delayedType->identifier().isExpression() ) {
        QList<QualifiedIdentifier> identifiers;
        identifiers << delayedType->identifier();
        QList<Declaration*> decls;
        
        if( !searchContext->findDeclarationsInternal( identifiers, searchContext->range().end, AbstractType::Ptr(), decls, inclusionTrace, searchFlags ) )
          return const_cast<AbstractType*>(type);
        
        if( !decls.isEmpty() )
          return applyPointerReference(decls.front()->abstractType().data(), delayedType->identifier());
      }
      ///Resolution as type has failed, or is not appropriate.
      ///Resolve delayed expression, for example static numeric expressions
      ExpressionParser p;
      ExpressionEvaluationResult res;
      if( delayedType->identifier().isExpression() )
        res = p.evaluateExpression( delayedType->identifier().toString().toUtf8(), DUContextPointer(const_cast<DUContext*>(searchContext)), inclusionTrace );
      else
        res = p.evaluateType( delayedType->identifier().toString().toUtf8(), DUContextPointer(const_cast<DUContext*>(searchContext)), inclusionTrace );

      ///@todo make this nicer
      keepAlive = res.type; //Since the API uses AbstractType*, use keepAlive to make sure the type cannot be deleted
      return res.type.data();
    }else{
      if( containsDelayedType(type) )
      {
        //Copy the type to keep the correct reference/pointer structure
        AbstractType* typeCopy = type->clone();
        PointerType* pType = dynamic_cast<PointerType*>(typeCopy);
        ReferenceType* rType = dynamic_cast<ReferenceType*>(typeCopy);
        if( pType ) //Replace the base
          pType->exchangeTypes(this);
        if( rType ) //Replace the base
          rType->exchangeTypes(this);
        
        return typeCopy;
      }
    }
    
    return const_cast<AbstractType*>(type); ///@todo remove const_cast;
  }

  virtual bool exchangeMembers() const {
    return false;
  }
  private:
    AbstractType::Ptr keepAlive;
};

AtomicIncrementer::Int DelayedTypeResolver::depth_counter = Q_BASIC_ATOMIC_INITIALIZER(0);

bool operator==( const ExpressionEvaluationResult& left, const ExpressionEvaluationResult& right ) {
 return left.type == right.type && left.instance.isInstance == right.instance.isInstance;
}

namespace Cpp {
// bool operator==( const QList<ExpressionEvaluationResult>& left, const QList<ExpressionEvaluationResult>& right ) {
//   return true;
// }

TemplateDeclaration::TemplateDeclaration(const TemplateDeclaration& /*rhs*/) : m_parameterContext(0), m_instantiatedFrom(0), m_specializedFrom(0) {
}

TemplateDeclaration::TemplateDeclaration() : m_parameterContext(0), m_instantiatedFrom(0), m_specializedFrom(0) {
}

TemplateDeclaration::~TemplateDeclaration()
{
  setSpecializedFrom(0);
  
  InstantiationsHash instantiations;
  {
    QMutexLocker l(&instantiationsMutex);
    instantiations = m_instantiations;

    ///Unregister at the declaration this one is instantiated from
    if( m_instantiatedFrom ) {
      InstantiationsHash::iterator it = m_instantiatedFrom->m_instantiations.find(m_instantiatedWith);
      if( it != m_instantiatedFrom->m_instantiations.end() && *it == this )
        m_instantiatedFrom->m_instantiations.erase(it);
      
      m_instantiatedFrom = 0;
    }
  }

  ///Delete all slave-declarations
  foreach( TemplateDeclaration* decl, instantiations ) {
    ///Specializations should always be anonymous, so we can delete them here.
    decl->m_instantiatedFrom = 0;
    delete decl;
  }

  QList<TemplateDeclaration*> specializations = m_specializations;
  foreach( TemplateDeclaration* specialization, specializations )
    specialization->setSpecializedFrom(0);
}

TemplateDeclaration* TemplateDeclaration::instantiatedFrom() const {
  return m_instantiatedFrom;
}

void TemplateDeclaration::setSpecializedFrom(TemplateDeclaration* other) {
  if( m_specializedFrom )
    m_specializedFrom->m_specializations.removeAll(this);

  m_specializedFrom = other;

  if( m_specializedFrom )
    m_specializedFrom->m_specializations << this;
}

TemplateDeclaration* TemplateDeclaration::specializedFrom() const {
  return m_specializedFrom;
}

void TemplateDeclaration::setInstantiatedFrom(TemplateDeclaration* from)
{
  Declaration* thisDecl = dynamic_cast<Declaration*>(this);
  //Change the identifier so it contains the template-parameters
  
  QMutexLocker l(&instantiationsMutex);
  if( m_instantiatedFrom ) {
    InstantiationsHash::iterator it = m_instantiatedFrom->m_instantiations.find(m_instantiatedWith);
    if( it != m_instantiatedFrom->m_instantiations.end() && *it == this )
      m_instantiatedFrom->m_instantiations.erase(it);

    m_instantiatedFrom = 0;
  }
  m_instantiatedFrom = from;
  from->m_instantiations.insert(m_instantiatedWith, this);
}

bool TemplateDeclaration::isInstantiatedFrom(const TemplateDeclaration* other) const {
    QMutexLocker l(&instantiationsMutex);
  
    InstantiationsHash::const_iterator it = other->m_instantiations.find(m_instantiatedWith);
    if( it != other->m_instantiations.end() && (*it) == this )
      return true;
    else
      return false;
}


void TemplateDeclaration::setTemplateParameterContext(KDevelop::DUContext* context) {
  m_parameterContext = context;
}

KDevelop::DUContext* TemplateDeclaration::templateParameterContext() const {
  return const_cast<KDevelop::DUContext*>(m_parameterContext.data()); ///@todo make data() const; return non-const pointer in duchain-pointer
}

bool isTemplateDeclaration(const KDevelop::Declaration* decl) {
  return (bool)dynamic_cast<const TemplateDeclaration*>(decl);
}

///Returns the context assigned to the given declaration that contains the template-parameters, if available. Else zero.
DUContext* getTemplateContext(Declaration* decl)
{
  DUContext* internal = decl->internalContext();
  if( !internal )
    return 0;
  foreach( DUContextPointer ctx, internal->importedParentContexts() ) {
    if( ctx )
      if( ctx->type() == DUContext::Template )
        return ctx.data();
  }
  return 0;
}

///Applies the default-parameters from basicDeclaration
void applyDefaultParameters( QList<Cpp::ExpressionEvaluationResult>& templateParameters, Declaration* basicDeclaration )
{
  DUContext* templateCtx = getTemplateContext(basicDeclaration);
  if( !templateCtx )
    return;

  if( templateParameters.count() >= templateCtx->localDeclarations().count() )
    return;
  
  for( int a = templateParameters.count(); a < templateCtx->localDeclarations().count(); a++ ) {
    TemplateParameterDeclaration* decl = dynamic_cast<TemplateParameterDeclaration*>(templateCtx->localDeclarations()[a]);
    if( decl ) {
      if( !decl->defaultParameter().isEmpty() ) {
        ExpressionEvaluationResult res;
        res.type = decl->abstractType();

        templateParameters << res;
      }else{
        //kDebug(9007) << "missing needed default template-parameter";
      }
    } else {
      kDebug(9007) << "Warning: non template-parameter in template context";
    }
  }
}

///@todo prevent endless recursion when resolving base-classes!(Parent is not yet in du-chain, so a base-class that references it will cause endless recursion)
CppDUContext<KDevelop::DUContext>* instantiateDeclarationContext( KDevelop::DUContext* parentContext, const ImportTrace& inclusionTrace, KDevelop::DUContext* context, const QList<ExpressionEvaluationResult>& _templateArguments, Declaration* instantiatedDeclaration, Declaration* instantiatedFrom )
{
  QList<ExpressionEvaluationResult> templateArguments = _templateArguments;

  if( templateArguments.count() == 1 && !templateArguments[0].isValid() )
    templateArguments.clear(); //If there is only exactly one template-argument, and that one is invalid, interpret the whole thing as no arguments at all
  
  if( instantiatedDeclaration )
    ///Move the instantiated declaration anonymously into the parent-context
    instantiatedDeclaration->setContext(parentContext, true);

  StandardCppDUContext* contextCopy = 0;
  
  if( context ) {
    ///Specialize involved contexts
    Q_ASSERT(context->parentContext()); //Top-context is not allowed
    contextCopy = new StandardCppDUContext(context->url(), context->range(), parentContext, true); //We do not need to care about TopDUContext here, because a top-context can not be instantiated
    contextCopy->setSmartRange(context->smartRange(), KDevelop::DocumentRangeObject::DontOwn); //The range belongs to the original context, so flag it not to be owned by the context
    contextCopy->setType(context->type());
    contextCopy->setLocalScopeIdentifier(context->localScopeIdentifier());

    if( instantiatedDeclaration )
      instantiatedDeclaration->setInternalContext(contextCopy);
      
    ///Now the created context is already partially functional and can be used for searching(not the instantiated template-params yet though)
    
    if( context->type() == KDevelop::DUContext::Template ) { //templateArguments may be empty, that means that only copying should happen.
      ///Specialize the local template-declarations
      QList<ExpressionEvaluationResult>::const_iterator currentArgument = templateArguments.begin();
      
      foreach(Declaration* decl, context->localDeclarations())
      {
        TemplateParameterDeclaration* templateDecl = dynamic_cast<TemplateParameterDeclaration*>(decl);
        Q_ASSERT(templateDecl); //Only template-parameter declarations are allowed in template-contexts
        TemplateParameterDeclaration* declCopy = dynamic_cast<TemplateParameterDeclaration*>(decl->clone());
        Q_ASSERT(declCopy);

        if( currentArgument != templateArguments.end() )
        {
          declCopy->setAbstractType( currentArgument->type );
          
          ++currentArgument;
        } else {
          //Use the already available delayed-type resolution to resolve the value/type
          if( !templateDecl->defaultParameter().isEmpty() ) {
            DelayedType::Ptr delayed( new DelayedType() );
            delayed->setIdentifier( templateDecl->defaultParameter() );
            declCopy->setAbstractType( resolveDelayedTypes( AbstractType::Ptr(delayed.data()), contextCopy, inclusionTrace + parentContext->topContext()->importTrace(contextCopy->topContext()) ) );
          }else{
            //Parameter missing
          }
        }
        ///This inserts the copied declaration into the copied context
        declCopy->setContext(contextCopy);
      }
    }
    
    foreach( KDevelop::DUContextPointer importedContext,  context->importedParentContexts() )
    {
      if( !importedContext)
        continue;
        ///For functions, the Template-context is one level deeper(it is imported by the function-context) so also copy the function-context
      if( importedContext->type() == KDevelop::DUContext::Template || importedContext->type() == KDevelop::DUContext::Function )
      {
        DUContext* ctx = instantiateDeclarationContext( parentContext, inclusionTrace, importedContext.data(), templateArguments, 0, 0);
        contextCopy->addImportedParentContext( ctx, SimpleCursor(), true );
        
        if( instantiatedDeclaration && importedContext->type() == KDevelop::DUContext::Template ) {
          TemplateDeclaration* tempDecl = dynamic_cast<TemplateDeclaration*>(instantiatedDeclaration);
          if( instantiatedDeclaration )
            tempDecl->setTemplateParameterContext( ctx );
          else
            kWarning(9007) << "instantiated declaration is not a template declaration";
        }
      }
      else
      {
        //Import all other imported contexts
        contextCopy->addImportedParentContext( importedContext.data(), SimpleCursor::invalid(), true );
      }
    }

    if( instantiatedDeclaration ) {
      ///We do not need to respect forward-declarations here, because they are not allowed as base-classes.
      const CppClassType* klass = dynamic_cast<const CppClassType*>( instantiatedDeclaration->abstractType().data() );
      if( klass ) { //It could also be a function
        ///Resolve template-dependent base-classes(They can not be found in the imports-list, because their type is DelayedType and those have no context)

        foreach( const CppClassType::BaseClassInstance& base, klass->baseClasses() ) {
          const DelayedType* delayed = dynamic_cast<const DelayedType*>(base.baseClass.data());
          if( delayed ) {
            ///Resolve the delayed type, and import the context
            DelayedTypeResolver res(contextCopy, inclusionTrace);
            AbstractType::Ptr newType( res.exchange(delayed) );

            if( CppClassType* baseClass = dynamic_cast<CppClassType*>(newType.data()) )
            {
              if( baseClass->declaration() && baseClass->declaration()->internalContext() )
              {
                contextCopy->addImportedParentContext( baseClass->declaration()->internalContext(), SimpleCursor::invalid(), true );
              }
            } else {
              kDebug(9007) << "Resolved bad base-class";
            }
          }
        }
      }
    }

  } else {
    ///Note: this is possible, for example for template function-declarations(They do not have an internal context, because they have no compound statement), for variables, etc..
    ///Directly take their assigned template-parameter-context and specialize it. We need it at least for overload-resolution.
    TemplateDeclaration* fromTemplateDecl = dynamic_cast<TemplateDeclaration*>(instantiatedFrom);
    TemplateDeclaration* toTemplateDecl = dynamic_cast<TemplateDeclaration*>(instantiatedDeclaration);
    if( toTemplateDecl && fromTemplateDecl && fromTemplateDecl->templateParameterContext() ) {
        DUContext* ctx = instantiateDeclarationContext( parentContext, inclusionTrace, fromTemplateDecl->templateParameterContext(), templateArguments, 0, 0);
        toTemplateDecl->setTemplateParameterContext( ctx );
    }
  }

  if( instantiatedDeclaration ) {
    TemplateDeclaration* instantiatedTemplate = dynamic_cast<TemplateDeclaration*>(instantiatedDeclaration);

    applyDefaultParameters( templateArguments, instantiatedDeclaration );
    
    if(instantiatedTemplate) //Since this is also called for normal members, this does not have to be the case.
      instantiatedTemplate->setInstantiatedWith(templateArguments);
      
    ///Change the identifier to reflect the set template-arguments
    if( !templateArguments.isEmpty() ) {
      KDevelop::Identifier id = instantiatedDeclaration->identifier();
      foreach(Cpp::ExpressionEvaluationResult expr, templateArguments)
        id.appendTemplateIdentifier(expr.identifier());

      instantiatedDeclaration->setIdentifier(id);
    }
  }  
  
  if( contextCopy )
    contextCopy->setInstantiatedFrom(dynamic_cast<CppDUContext<DUContext>*>(context), templateArguments);
  ///Since now the context is accessible through the du-chain, so it must not be changed any more.

  if( instantiatedDeclaration && instantiatedDeclaration->abstractType() ) {

    ///Resolve all involved delayed types
    
    IdentifiedType* idType = dynamic_cast<IdentifiedType*>(instantiatedDeclaration->abstractType().data());

    ///Use the internal context if it exists, so undefined template-arguments can be found and the DelayedType can be further delayed then.
    AbstractType::Ptr changedType = resolveDelayedTypes( instantiatedDeclaration->abstractType(), instantiatedDeclaration->internalContext() ? instantiatedDeclaration->internalContext() : parentContext, inclusionTrace );

    if( idType && idType->declaration() == instantiatedFrom ) {
      if( changedType == instantiatedDeclaration->abstractType() )
          changedType = instantiatedDeclaration->abstractType()->clone();
      
      IdentifiedType* changedIdType = dynamic_cast<IdentifiedType*>(changedType.data());
      if( changedIdType )
        changedIdType->setDeclaration(instantiatedDeclaration);
    }

    instantiatedDeclaration->setAbstractType( changedType );
  }
  
  return contextCopy;
}


///@todo Use explicitly declared specializations
///@todo Implement correct instantiation-cache updating in templatedeclaration.cpp, which includes correct checking for dead forward-declarations and any other dead declarations
Declaration* TemplateDeclaration::instantiate( const QList<ExpressionEvaluationResult>& templateArguments, const ImportTrace& inclusionTrace )
{
  if( m_instantiatedFrom )
    return m_instantiatedFrom->instantiate( templateArguments, inclusionTrace ); ///@todo update inclusion-trace
  
  {
    QMutexLocker l(&instantiationsMutex);
   InstantiationsHash::const_iterator it;
    it = m_instantiations.find( InstantiationKey(templateArguments) );
    if( it != m_instantiations.end() ) {
      return dynamic_cast<Declaration*>(*it);
    }
  }

  
  Declaration* decl = dynamic_cast<Declaration*>(this);
  Q_ASSERT(decl);
  Declaration* clone = decl->clone();
  Q_ASSERT(clone);
  TemplateDeclaration* cloneTemplateDecl = dynamic_cast<TemplateDeclaration*>(clone);
  Q_ASSERT(cloneTemplateDecl);
  
  ///Now eventually create the virtual contexts
  instantiateDeclarationContext( decl->context(), inclusionTrace, decl->internalContext(), templateArguments, clone, decl );

  cloneTemplateDecl->setInstantiatedFrom(this);

  return clone;
}

AbstractType::Ptr resolveDelayedTypes( AbstractType::Ptr type, const KDevelop::DUContext* context, const KDevelop::ImportTrace& inclusionTrace, KDevelop::DUContext::SearchFlags searchFlags ) {
  if( !type )
    return type;

  ///First, find out if delayed types are involved(Function return-type, function-arguments, reference/pointer target, base-class)
  DelayedTypeSearcher search;
  
  type->accept(&search);

  DelayedType* delayedType = dynamic_cast<DelayedType*>(type.data());

  if( search.found || delayedType ) {
    ///Delayed types were found. We must copy the whole type, and replace the delayed types.

    DelayedTypeResolver resolver(context, inclusionTrace, searchFlags);

    AbstractType::Ptr typeCopy;
    if( delayedType )
      ///The type itself is a delayed type, resolve it
      typeCopy = resolver.exchange( type.data() );
    else {
      ///Resolve involved delayed types, now hopefully we know the template-parameters
      typeCopy = AbstractType::Ptr( type->clone() );
      DelayedTypeSearcher testSearch;
      typeCopy->accept(&testSearch);
      Q_ASSERT(testSearch.found);
      typeCopy->exchangeTypes( &resolver );
    }

    return typeCopy;
  } else {
    return type;
  }
}

void TemplateDeclaration::setInstantiatedWith(const QList<ExpressionEvaluationResult>& with) {
  Q_ASSERT(!m_instantiatedFrom);
  m_instantiatedWith.args = with;
}

const QList<ExpressionEvaluationResult>& TemplateDeclaration::instantiatedWith() const {
  return m_instantiatedWith.args;
}

TemplateDeclaration::InstantiationsHash TemplateDeclaration::instantiations() const {
    QMutexLocker l(&instantiationsMutex);
    return m_instantiations;
}

template<>
Declaration* SpecialTemplateDeclaration<ForwardDeclaration>::resolve(const TopDUContext* topContext) const {
  if( instantiatedFrom() ) {
    SpecialTemplateDeclaration<ForwardDeclaration>* instantiatedFrom = dynamic_cast<SpecialTemplateDeclaration<ForwardDeclaration>*>(this->instantiatedFrom());
    if( instantiatedFrom ) {
      Declaration* baseResolved = instantiatedFrom->resolve(topContext);
      TemplateDeclaration* baseTemplate = dynamic_cast<TemplateDeclaration*>(baseResolved);
      if( baseResolved && baseTemplate ) {
        Declaration* ret = baseTemplate->instantiate(instantiatedWith(), topContext ? topContext->importTrace(this->topContext()) : ImportTrace());
        return ret;
      }else{
          //Forward-declaration was not resolved
          return 0;
      }
    }else{
      kWarning(9007) << "Problem in template forward-declaration";
      return 0;
    }
  }else{
    return ForwardDeclaration::resolve(topContext);
  }
}

}

